
//
//
// Project       : Vorpal-Mail
//
// Source        : session.h
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 5th December 2000
//
// License       : Copyright Virtua-Web (GPL to be used once release 1.0 is achieved).
//
// Description   :
//
// Definition files for SMTP session class. Supports a single SMTP session, parsing
// the relevant commands to provide an RFC 821 compliant session.
//
// Changelog     :
//

#include "smtp.h"
#include "virtual.h"

#ifndef __SMTP_SESSION__
#define __SMTP_SESSION__

enum SMTP_CODE {
	SMTP_ERROR = -1,
	SMTP_OK = 0,
	};

enum SMTP_MODE {
	SMTP_MODE_NONE = 0,
	SMTP_MODE_SERVER,
	SMTP_MODE_CLIENT,
	};

enum SMTP_STATUS {
	SMTP_STATUS_OPEN = 0,

	SMTP_STATUS_HELO,		// awaiting HELO
	SMTP_STATUS_MAIL,		// got HELO, awaiting MAIL (start of msg)
	SMTP_STATUS_RCPT,		// got MAIL, awaiting RCPT or DATA
	SMTP_STATUS_DATA,		// in DATA, waiting for .
	SMTP_STATUS_RSET,

	SMTP_STATUS_DATATEXT,
	SMTP_STATUS_SENT,		// client only, msg sent
	SMTP_STATUS_DATASEND,		// client only, still sending data
	SMTP_STATUS_SENDRETRY,		// client only, didnt buffer, so wait and retry

	SMTP_STATUS_COMPLETE = 100,

	SMTP_STATUS_TIMEOUT = -2,	// used to process timeouts

	SMTP_STATUS_ABORT = -1,
	

	};

enum {
	SMTP_CMD_QUIT = 0,
	SMTP_CMD_HELP,
	SMTP_CMD_HELO,
	SMTP_CMD_EHLO,
	SMTP_CMD_MAIL,
	SMTP_CMD_RCPT,
	SMTP_CMD_DATA,
	SMTP_CMD_RSET,
	SMTP_CMD_NOOP,
	SMTP_CMD_VRFY,
	SMTP_CMD_EXPN,
	};

typedef struct SMTP_SESSION {
	private:
		static char *commands[];
		static char *command_help[];
		static char *command_arg[];
		static char *hostname;
		static long sequence;
		static long pid;
		static int debug;
		static VIRTUAL_DOMAIN *domain;

		SMTP_MODE mode;

		char last_response[SMTP_MAX_LINE];
		
		char *filename;
		FILE *bodyfile, *bouncefile, *envelopefile;
		int socket;
		SMTP_STATUS status;
		char *line;
		char *scratch;
		char *spill;
		SMTP_SESSION *next, *prev;
		struct sockaddr_in saddr;

		LIST *sendbuffer;
		int sendcount;

		char *from;
		STRING_ARRAY *to;

		int loop, relayok;
		int dataheader, datamime;
		char *datamimesection;
		int bounce, bouncing;

		long qtime, ltime;

		int timeout;

		void initialise(void);
		int sendf(char *, ...);
		int lookup(char *);
		void process_server(char *);
		void process_client(char *);
		void process_cmd(char *);
		int parsename(char *, char *, char *);
		void addreceive(void);
		char *sessionname(void);
		char *makefilename(char *, char, char *);
		int checkthisrelay(char *, char *);
		int checkthislist(char *, LIST *);
		void localdeliver(void);
		int dolocaldeliver(char *, char *, VIRTUAL_DOMAIN *);
		int localuser(char *, VIRTUAL_DOMAIN *);
		void doremotedeliver(void);
		void remotedeliver(void);
		void addbounce(char *, ...);
		void sendbounce(void);
		void errormail(char *);
		void cleanup(char *);
		void closefiles(void);
		int lock(FILE *);
		void unlock(FILE *);
		void virushdr(char *, char *);
		void virusmsg(char *);
		int scanvirus(char *);
		void addheader(char *, ...);
		int splitheader(char *, char *, char *);
		void process_data(char *);
	public:
		SMTP_SESSION(void);
		SMTP_SESSION(int, struct sockaddr_in *);
		~SMTP_SESSION();
		int openfiles(char *);
		char *getfilename(void) { return filename; };
		SMTP_STATUS getstatus(void) { return status; };
		int getsocket(void) { return socket; };
		void setsocket(int s) { socket = s; };
		void setnext(SMTP_SESSION *);
		SMTP_SESSION *getnext(void) { return next; };
		SMTP_SESSION *getprev(void) { return prev; };
		int setmode(SMTP_MODE);
		void process(char *);
		void closesocket(void);
		void serverstatus(int);
		void serverstatus(int, char *);
		void clientstatus(SMTP_STATUS);
		void clientstatus(SMTP_STATUS, char *);
		void setfrom(char *);
		void addto(char *);
		void setdebug(int n) { debug = n; };
		void setdomain(VIRTUAL_DOMAIN *);
		void writeheader(void);
		int checkrelay(char *);
		void pulse(void);
		void reset(void);
		int pending(void);
		void dosend(void);
		void sendreset(void);
		void sendquit(void);
		void setqtime(long t) { qtime = t; };
		void setltime(long t) { ltime = t; };
		char *getresponse(void) { return last_response; };
	};


#endif

