
//
//
// Project       : Vorpal-Mail
//
// Source        : sessions.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 7th December 2000
//
// License       : Copyright Virtua-Web (GPL to be used once release 1.0 is achieved)
//
// Description   :
//
// Class implementation for SMTP session. Implements the server side of an RFC821 compliant
// SMTP server.
//
// Changelog     :
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <syslog.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <pwd.h>

#include "mail.h"
#include "array.h"
#include "session.h"
#include "virtual.h"
#include "smtp.h"

//
// Relaying should be enabled under the following circumstances:
//
// the from username matches one in the relaylist	(user@domain)
// the from domain matches a domain in the relaylist	(@domain)
// the remote host matches a hostname in the relaylist	(domain)
// the remote host matches an ip in the relaylist	(w.x.y.z)
//

int SMTP_SESSION::checkthisrelay(char *from, char *rule)
{	
if (strchr(rule, '@') != NULL)		// it's an email relay
	{
	if (*rule == '@')		// anyone from the domain
		{
		if (strcasecmp(rule, strchr(from, '@')) == 0)
			return 1;
		else
			return 0;
		}
	else if (strcasecmp(from, rule) == 0)	// specific
		return 1;
	else
		return 0;
	}
else					// it's a host/domain relay
	{
	struct hostent *host;
	host = gethostbyaddr((char *)&saddr.sin_addr.s_addr, sizeof(saddr.sin_addr.s_addr), AF_INET);
	if (strchr(rule, '*') != NULL)	// contains wildcards
		{
		char *pfrom, *pdom;
		pfrom = host->h_name; 		// check the CONNECTED hostname
		pdom = rule;
		while(*pfrom != 0 && *pdom != 0)
			{
			if (*pdom == '*')
				{
				pdom++;
				while(*pfrom != *pdom)
					{
					pfrom++;
					if (*pfrom == 0)
						return 0;
					}
				}
			else if (*pdom == *pfrom)
				{
				pdom++;
				pfrom++;
				}
			else
				return 0;
			}
		if (*pfrom == *pdom)
			return 1;
		else
			return 0;
		}
	else if (strchr(rule, '/') != NULL)	// it's a subnet
		{
		char *subnet, *ptr;
		int bmask;
		subnet = strdup(rule);
		if ((ptr = strchr(subnet, '/')) != NULL)
			{
			int a,b,c,d;
			*ptr = 0;
			bmask = strtol(ptr+1, NULL, 10);
			if (bmask <= 32 && bmask >= 0 && sscanf(subnet, "%d.%d.%d.%d", &a, &b, &c, &d) == 4)
				{
				unsigned long addr1, addr2, mask = 0l;
				addr1 = (a << 24) + (b << 16) + (c << 8) + d;
				addr2 =  ntohl(saddr.sin_addr.s_addr);
				while(bmask-- > 0)
					{
					mask >>= 1;
					mask |= 0x80000000;
					}
				if ((addr1 & mask) == (addr2 & mask))
					return 1;
				}
			}
		return 0;
		}
	else				// it's a host (could be IP or domainname)
		{
		char *rhost;
		rhost = strdup(host->h_name);			// copy the name of remote host
		if ((host = gethostbyname(rule)) != NULL)	// get the hostname specified in the rule
			{
			if (strcasecmp(host->h_name, rhost) == 0)
				{
				free(rhost);
				return 1;
				}
			}
		free(rhost);
		return 0;
		}
	}
}


int SMTP_SESSION::checkthislist(char *from, LIST *sublist)
{
LISTITEM *item;

item = sublist->gethead();
while(item != NULL)
	{
	if (!checkthisrelay(from, (char *)item->getkey()))
		return 0;
	item = item->getnext();
	}
return 1;
}

int SMTP_SESSION::checkrelay(char *from)
{
int n = 0;
VIRTUAL_DOMAIN *dom;
char *ptr;
LISTITEM *item;
LIST *list;

ptr = strchr(from, '@');

if (ptr == NULL || (dom = domain->finddomain(ptr+1)) == NULL)
	dom = domain;
	
if ((list = dom->getrelay()) != NULL)
	{
	item = list->gethead();
	while(item != NULL)
		{
		if (checkthislist(from, (LIST *)item->getdata()))
			return 1;
		item = item->getnext();
		}
	}
return 0;
}


void SMTP_SESSION::process_server(char *data)		// very simple state machine!
{
int cmd;
char word[SMTP_MAX_LINE],*from,*to;

from = data;
to = word;

if (status == SMTP_STATUS_DATA)
	{
	process_data(data);
	return;
	}

if (SMTP_SESSION::debug > 1)
	syslog(LOG_MAIL + LOG_DEBUG, "<< %s\n", data);

while(*from == ' ')
	from++;

while(*from != ' ' && *from != 0)			// parse out the first token on the command line
	*to++ = *from++;
*to = 0;

while(*from == ' ')
	from++;

this->line = from;

switch((cmd = lookup(word)))				// lookup the command, some commands work irrespective of state
	{
	case -1:					// not a command at all!
		this->line = data;
		serverstatus(500, this->line);
		return;
	case SMTP_CMD_RSET:
		status = SMTP_STATUS_RSET;
		serverstatus(250);
		closefiles();
		cleanup(NULL);
		reset();
		status = SMTP_STATUS_MAIL;
		return;
	case SMTP_CMD_HELP:				// mainly HELP
		{
		int n = 0;
		sendf("214-Welcome to " PROGRAM " " VERSION "\r\n");
		sendf("214-The following commands are supported:\r\n");
		while(SMTP_SESSION::commands[n] != NULL)
			{
			sendf("214-%s %-20s %s\r\n", SMTP_SESSION::commands[n], SMTP_SESSION::command_arg[n], SMTP_SESSION::command_help[n]);
			n++;
			}
		sendf("214-\r\n");
		sendf("214-To report bugs, please email bugs@virtua-web.co.uk\r\n");
		sendf("214 [End of help]\r\n");
		}
		return;
	case SMTP_CMD_QUIT:
		serverstatus(221);
		closesocket();				// abort session!
		this->status = SMTP_STATUS_ABORT;
		return;
	case SMTP_CMD_VRFY:		// 250 <user>
		{
		LIST *expanded;
		if (SMTP_SESSION::domain != NULL && (expanded = domain->expand(this->line)) != NULL)
			{
			if (strchr(this->line, '@') == NULL)
				sendf("250 <%s@%s>\r\n", this->line, domain->getname());
			else
				sendf("250 <%s>\r\n", this->line);
			delete expanded;
			}
		else
			sendf("550 <%s> unknown\r\n", this->line);
		}
		return;
	case SMTP_CMD_EXPN:		// 550 <user> unknown
		{
		LIST *expanded;
		if (SMTP_SESSION::domain != NULL && (expanded = domain->expand(this->line)) != NULL)
			{
			LISTITEM *item;
			item = expanded->gethead();
			while(item != NULL)
				{
				sendf("250%c<%s>\r\n", (item->getnext() == NULL) ? ' ' : '-', item->getkey());
				item = item->getnext();
				}
			delete expanded;
			}
		else
			sendf("550 <%s> unknown\r\n", this->line);
		}
		return;
	}

switch(status)
	{
	case SMTP_STATUS_HELO:		/* expecting EHLO or HELO */
		switch(cmd)
			{
			case SMTP_CMD_HELO:
			case SMTP_CMD_EHLO:
				serverstatus(250, this->line);
				status = SMTP_STATUS_MAIL;
				break;
			default:
				serverstatus(503, data);
				break;
			}
		break;
	case SMTP_STATUS_MAIL:
	case SMTP_STATUS_COMPLETE:
		switch(cmd)
			{
			case SMTP_CMD_MAIL:
				if (!parsename(line, "FROM:", word))
					serverstatus(501, line);
				else
					{
					this->closefiles();
					this->cleanup(NULL);
					this->reset();			// new message, so auto reset data
					status = SMTP_STATUS_MAIL;
					setfrom(word);
					if (openfiles(sessionname()))		// create a session
						{
						syslog(LOG_MAIL + LOG_NOTICE, "%s: MAIL FROM %s", filename, word);
						addreceive();
						serverstatus(250, word);
						status = SMTP_STATUS_RCPT;
						}
					else
						{
						serverstatus(451);
						closefiles();	
						cleanup(NULL);
						reset();
						status = SMTP_STATUS_MAIL;
						}
					}
				break;
			default:
				serverstatus(503, data);
				break;
			}
		break;
	case SMTP_STATUS_RCPT:
		switch(cmd)
			{
			case SMTP_CMD_RCPT:
				if (!parsename(line, "TO:", word))
					serverstatus(501, line);
				else
					{
					LIST *local = NULL;
					if (domain != NULL && (local = domain->expand(word)) != NULL)
						{
						LISTITEM *item;
						item = local->gethead();
						while(item != NULL)
							{
							addto(item->getkey());
							item = item->getnext();
							}
						delete local;
						serverstatus(250, word);
						}
					else
						{
						if (this->relayok)
							{
							if (SMTP_SESSION::debug)
								syslog(LOG_MAIL + LOG_NOTICE, "%s: RCPT TO %s accepted", filename, word);
							addto(word);
							serverstatus(250, word);
							}
						else
							{
							syslog(LOG_MAIL + LOG_NOTICE, "%s: RCPT TO %s REJECTED", filename, word);
							serverstatus(550, word);
							}
						}
					}

				break;
			case SMTP_CMD_DATA:
				serverstatus(354);
				status = SMTP_STATUS_DATA;
				break;
			default:
				serverstatus(503, data);
				break;
			}
		break;
	}
}


void SMTP_SESSION::serverstatus(int stat)
{
serverstatus(stat, "[none]");
}

void SMTP_SESSION::serverstatus(int stat, char *str)
{
switch(stat)
	{
	case SMTP_STATUS_TIMEOUT:
		closefiles();
		cleanup(NULL);
		reset();
		status = SMTP_STATUS_MAIL;
		break;
	case 500:
		sendf("500 Syntax error, command not recognised (%s)\r\n", str);
		break;
	case 501:	
		sendf("501 Syntax error in parameter: %s\r\n", str);
		break;
	case 503:
		sendf("503 Bad sequence of commands\r\n");
		break;
	case 550:
		sendf("550 <%s> relaying denied.\r\n", str);
		break;
	case 451:
		sendf("451 INTERNAL ERROR. MAIL CANNOT BE SENT.\r\n");
		break;
	case 354:
		sendf("354 Start mail input. Enter a period on a line by itself to end.\r\n");
		break;
	case 250:		// acknowledge command
		switch(this->status)
			{
			case SMTP_STATUS_HELO:			// initial connection
				{
				struct hostent *host;
				host = gethostbyaddr((char *)&saddr.sin_addr.s_addr, sizeof(saddr.sin_addr.s_addr), AF_INET);
				syslog(LOG_MAIL + LOG_NOTICE, "HELO %s [%s]", str, host->h_name);
				sendf("250 Welcome %s [%s].\r\n", str, host->h_name);
				}
				break;
			case SMTP_STATUS_RSET:
				sendf("250 Mail session reset.\r\n");
				break;
			case SMTP_STATUS_MAIL:
				sendf("250 Mail from %s accepted (%s).\r\n", str, (relayok ? "relaying enabled" : "relaying disabled"));
				break;
			case SMTP_STATUS_RCPT:
				sendf("250 Recipient %s added.\r\n", str);
				break;
			case SMTP_STATUS_DATA:
				sendf("250 %s Message accepted for delivery.\r\n", filename);
				this->status = SMTP_STATUS_COMPLETE;
				break;
			default:
				sendf("250 Request completed, OK.\r\n");
				break;
			}
		break;
	case 220:
		{
		time_t t;
		char tm[128];

		time(&t);
		strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));
		sendf("220 %s SMTP " PROGRAM " " VERSION " %s\r\n", SMTP_SESSION::hostname, tm);
		}
		break;
	case 221:
		sendf("221 %s closing connection. Bye.\r\n", SMTP_SESSION::hostname);
		closesocket();
		break;
	}
}
